#!/opt/opensvc/bin/python
""" 
Verify crontab content. Fix if appropriate.
The collector provides the format with wildcards.
The module replace the wildcards with contextual values.

The variable format is json-serialized, ex:

{
  "user": "sys",
  "check": "/opt/opensvc/bin/svcmon",
  "entry": "3,13,23,33,43,53 * * * *  /opt/opensvc/bin/svcmon --updatedb --maxdelaydb 120 >/dev/null 2>&1"
 - or - 
  "ref": "http:// ..."
}

Wildcards:
%%ENV:VARNAME%%		Any environment variable value
%%HOSTNAME%%		Hostname
%%SHORT_HOSTNAME%%	Short hostname

"""

import os
import sys
import json
import stat
import re
import urllib
import tempfile
import pwd
import grp
from subprocess import *

sys.path.append(os.path.dirname(__file__))

from comp import *

class CrontabEntry(object):
    def __init__(self, prefix='OSVC_COMP_CRONTABENTRY_'):
        self.prefix = prefix.upper()
        self.crontabs = {}
        self.checks = []
        self.upds = {}

        self.sysname, self.nodename, x, x, self.machine = os.uname()

        for k in [ key for key in os.environ if key.startswith(self.prefix)]:
            try:
                self.add_crontab(os.environ[k])
            except ValueError:
                print >>sys.stderr, 'key syntax error on var[', k, '] = ',os.environ[k]

        if len(self.checks) == 0:
            print "no appliable variable found in rulesets", self.prefix
            raise NotApplicable()

    def fixable(self):
        return RET_OK

    def get_env_item(self, d, item):
        if item not in d:
            return d
        if type(d[item]) != str and type(d[item]) != unicode:
            return d
        if not d[item].startswith("%%"):
            return d
        s = d[item].strip('%').replace('ENV:', '')
        if not s.startswith('OSVC_COMP_'):
            s = 'OSVC_COMP_'+s
        if s not in os.environ:
            print >>sys.stderr, "%s is not set"%s
            raise
        d[item] = os.environ[s]
        try:
            d[item] = int(d[item])
        except:
            pass
        return d

    def parse_entry(self, x):
        if isinstance(x, int):
            x = str(x)+'\n'
        x = x.replace('%%HOSTNAME%%', self.nodename)
        x = x.replace('%%SHORT_HOSTNAME%%', self.nodename.split('.')[0])
        if not x.endswith('\n'):
            x += '\n'
        return x

    def parse_ref(self, url):
        f = tempfile.NamedTemporaryFile()
        tmpf = f.name
        try:
            fname, headers = urllib.urlretrieve(url, tmpf)
            if 'invalid file' in headers.values():
                print >>sys.stderr, url, "not found on collector"
                return RET_ERR
            content = unicode(f.read())
            f.close()
        except:
             print >>sys.stderr, url, "not found on collector"
             return ''
        if '<title>404 Not Found</title>' in content:
            print >>sys.stderr, url, "not found on collector"
            return ''
        return self.parse_entry(content)

    def subst(self, v):
        if type(v) == list:
            l = []
            for _v in v:
                l.append(self.subst(_v))
            return l
        if type(v) != str and type(v) != unicode:
            return v

        p = re.compile('%%ENV:\w+%%')
        for m in p.findall(v):
            s = m.strip("%").replace('ENV:', '')
            if s in os.environ:
                _v = os.environ[s]
            elif 'OSVC_COMP_'+s in os.environ:
                _v = os.environ['OSVC_COMP_'+s]
            else:
                print >>sys.stderr, s, 'is not an env variable'
                raise NotApplicable()
            v = v.replace(m, _v)
        return v

    def read_crontab(self, user):
        if user == '':
            user = 'root'
        cmd = ['/usr/bin/crontab', '-l', user]
        p = Popen(cmd, stdout=PIPE, stderr=PIPE)
        out,err = p.communicate()
        if p.returncode != 0 :
            print >>sys.stderr, "Cannot get %s's %s" %(user,err.strip('\n'))
            return RET_ERR, ''
        return RET_OK, out

    def add_crontab(self,v):
        r = RET_OK
        d = json.loads(v)
        if 'user' not in d:
            print >>sys.stderr, 'user should be defined:', d
            r |= RET_ERR
        if 'entry' in d and 'ref' in d:
            print >>sys.stderr, 'entry and ref are exclusive:', d
            r |= RET_ERR
        try:
            d = self.get_env_item(d, 'check')
        except:
            r |= RET_ERR
        for k in ('user', 'entry', 'check', 'ref'):
            if k in d:
                d[k] = self.subst(d[k])
        if len(d['user']) < 1:
            d['user'] = 'root'
        if not d['user'] in self.upds:
            self.upds[d['user']] = 0
        if not d['user'] in self.crontabs:
            rx,text = self.read_crontab(d['user'])
            if rx == RET_OK:
                self.crontabs[d['user']] = text
            else:
                self.crontabs[d['user']] = ''
        c = ''
        if 'entry' in d:
            c = self.parse_entry(d['entry'])
        elif 'ref' in d:
            c = self.parse_ref(d['ref'])
        else:
            print >>sys.stderr, 'entry or ref should be defined:', d
            r |= RET_ERR
        if d['check'] in c:
            val = True
        else:
            val = False
            r |= RET_ERR
        self.checks.append({'user':d['user'], 'check':d['check'], 'add':c, 'valid':val})
        return r
        
    def check_crontab(self):
        r = RET_OK
        for ck in self.checks:
             if not ck['valid']:
                 print >>sys.stderr, "Pattern '%s' for %s's crontab is absent in the requested content" %(ck['check'], ck['user'])
                 r |= RET_ERR
                 continue
             pr = RET_ERR
             lines = self.crontabs[ck['user']].split('\n')
             for line in lines:
                 if line.startswith('#'):
                     continue
                 if ck['check'] in line:
                     pr = RET_OK
                     break
             if pr == RET_OK:
                 print "Pattern '%s' matches %s's crontab, entry: %s => OK" %(ck['check'], ck['user'], '"'+line+'"')
             else:
                 print >>sys.stderr, "Pattern '%s' **does not match** %s's crontab => BAD" %(ck['check'], ck['user'])
             r |= pr
        return r

    def check(self):
        r = 0
        r |= self.check_crontab()
        return r

    def add_2_crontabs(self):
        r = RET_OK
        f = tempfile.NamedTemporaryFile()
        filen = f.name
        f.close()
        for user in self.crontabs:
            if self.upds[user] == 0:
                continue
            f = open(filen, 'w+')
            try:
                f.writelines(self.crontabs[user])
                f.close()
            except:
                print >>sys.stderr, "fail to close temporary file %s" %filen
                r |= RET_ERR
            if user != 'root':
                cmd = ['/usr/bin/su', user, '-c', '/usr/bin/crontab '+filen]
            else:
                cmd = ['/usr/bin/crontab', filen]
            p = Popen(cmd, stdout=PIPE, stderr=PIPE)
            out,err = p.communicate()
            if p.returncode != 0:
                print >>sys.stderr, "Could not append-to or create %s's %s" %(user,err.strip('\n'))
                r |= RET_ERR
            else:
                print "Success: %s's crontab => OK" %user
            try:
                os.unlink(filen)
            except:
                print >>sys.stderr, 'Could not remove temp file: %s' %filen
                pass
        return r

    def fix_crontabs(self):
        r = RET_OK
        for ck in self.checks:
             if not ck['valid']:
                 print >>sys.stderr, "Pattern '%s' for %s's crontab is absent in the requested content" %(ck['check'], ck['user'])
                 r |= RET_ERR
                 continue
             pr = RET_ERR
             lines = self.crontabs[ck['user']].split('\n')
             for line in lines:
                 if line.startswith('#'):
                     continue
                 if ck['check'] in line:
                     pr = RET_OK
                     break
             if pr != RET_OK:
                 print "Trying to add to %s's crontab, entry: %s" %(ck['user'], ck['add'].strip('\n'))
                 self.crontabs[ck['user']] += ck['add']
                 self.upds[ck['user']] = 1
             r |= pr
        r |= self.add_2_crontabs()
        return r

    def fix(self):
        r = 0
        r != self.fix_crontabs()
        return r


if __name__ == "__main__":
    syntax = """syntax:
      %s PREFIX check|fixable|fix"""%sys.argv[0]
    if len(sys.argv) != 3:
        print >>sys.stderr, "wrong number of arguments"
        print >>sys.stderr, syntax
        sys.exit(RET_ERR)
    try:
        o = CrontabEntry(sys.argv[1])
        if sys.argv[2] == 'check':
            RET = o.check()
        elif sys.argv[2] == 'fix':
            RET = o.fix()
        elif sys.argv[2] == 'fixable':
            RET = o.fixable()
        else:
            print >>sys.stderr, "unsupported argument '%s'"%sys.argv[2]
            print >>sys.stderr, syntax
            RET = RET_ERR
    except ComplianceError:
        sys.exit(RET_ERR)
    except NotApplicable:
        sys.exit(RET_NA)
    except:
        import traceback
        traceback.print_exc()
        sys.exit(RET_ERR)

    sys.exit(RET)

